# -*- coding: utf-8 -*-

import os
import sys
import argparse 

from diwork_ways import *

def main_repeats(args: list):
    parser = argparse.ArgumentParser(prog = "diwork repeats",
        description="This module will find duplicates (exactly the same files) in the directory")
    parser.add_argument("folder_path", type=str, nargs=1,
                       help="Path to directory where repeats will be found")
    parser.add_argument("--delete", default=False, action='store_true',
                       help="Deletes files so that only 1 duplicate remains")
    parser.add_argument("--yes", default=False, action='store_true',
                       help="No interractive. Answer \"yes\" always. ")
    parser = common_init_parser(parser)
    args = parser.parse_args(args)
    common_init_parse(args)

    folder = args.folder_path[0]
    IF_DELETE = args.delete
    IF_YES = args.yes
    err_out = []
    files_to_delete = []
    folder_abs = os.path.abspath(folder)
    if(is_folder(folder_abs) == False):
        pout(f"\"{folder_abs}\" is not folder. ")
        exit()
    files_abs = sorted(get_files_list(folder_abs))
    hashes = set()
    d = {}
    gi, N = 0, len(files_abs)
    for file_i in files_abs:
        gi+=1
        if(is_file(file_i) == False):
            err_out.append(f"\"{file_i}\" is not file or does not exists, it will be skipped. ")
            continue
        file_i_hash = get_hash_file(file_i)
        pout(f"({gi}/{N}) Calculated hash of \"{file_i}\": {file_i_hash}")
        if(file_i_hash not in hashes):
            hashes.add(file_i_hash)
            d[file_i_hash] = [file_i]
        else:
            d[file_i_hash].append(file_i)

    pout("\n===============\nIdentical files: ")
    IF_AT_LEAST_ONE = False
    hashesss = list(d.keys())
    for hash_i in hashesss:
        fl = d[hash_i]
        if(len(fl) > 1):
            IF_AT_LEAST_ONE = True
            pout(f"* Hash \"{hash_i}\" have files: ")
            li = 0
            for file_i in fl:
                if(li > 0 and IF_DELETE):
                    pout(f"\t{file_i} (will be deleted)")
                    files_to_delete.append(file_i)
                else:
                    pout(f"\t{file_i}")
                li-=-1
            pout("")

    if(IF_AT_LEAST_ONE == False):
        pout("\tNo such files")
    else:
        pout("\nThis file will be delete:")
        for del_file_i in files_to_delete:
            pout(f"\"{del_file_i}\"")
        if(IF_YES == False):
            input(f"Type enter to continue or press CTRL+C to cancel.")
        for del_file_i in files_to_delete:
            os.remove(del_file_i)
        pout("DELETED: OK")



    if(len(err_out) != 0):
        pout(f"\n===============\nSome troubles happened:")
        for err_i in err_out:
            pout(f"\t{err_i}")
        pout(f"===============")

    pout("=============== Done! ===============")



# -*- coding: utf-8 -*-

import os
import sys
import argparse 

from diwork_ways import *

def main_difx(args: list):
    argc = len(args)
    if(argc != 2):
        pout("This module will show the difference between the two directories.\n")
        pout("Syntax error. Expected: \"python folder_work.py difx {folder1} {folder2}\"")
        exit()
    folder1 = args[0]
    folder2 = args[1]
    err_out = []
    folder1_abs = os.path.abspath(folder1)
    folder2_abs = os.path.abspath(folder2)
    if(is_folder(folder1_abs) == False):
        pout(f"\"{folder1_abs}\" is not folder. ")
        exit()
    if(is_folder(folder2_abs) == False):
        pout(f"\"{folder2_abs}\" is not folder. ")
        exit()

    files1_abs, files2_abs = sorted(get_files_list(folder1_abs)), sorted(get_files_list(folder2_abs))
    dr1, dr2 = {}, {}
    gi, N = 0, len(files1_abs)+len(files2_abs)
    for file_i in files1_abs:
        gi+=1
        if(is_file(file_i) == False):
            err_out.append(f"\"{file_i}\" is not file or does not exists, it will be skipped. ")
            continue
        file_i_hash = get_hash_file(file_i)
        pout(f"({gi}/{N}) Calculated hash of \"{file_i}\": {file_i_hash}")
        file_i_rel = rel_path(file_i, folder1_abs)
        dr1[file_i_rel] = file_i_hash
    for file_i in files2_abs:
        gi+=1
        if(is_file(file_i) == False):
            err_out.append(f"\"{file_i}\" is not file or does not exists, it will be skipped. ")
            continue
        file_i_hash = get_hash_file(file_i)
        pout(f"({gi}/{N}) Calculated hash of \"{file_i}\": {file_i_hash}")
        file_i_rel = rel_path(file_i, folder2_abs)
        dr2[file_i_rel] = file_i_hash
    files1_rel, files2_rel = list(dr1.keys()), list(dr2.keys())
    files1_rel_set, files2_rel_set = set(dr1.keys()), set(dr2.keys())

    pout("\n===============\nDifx:")

    IF_AT_LEAST_ONE = False

    for file_i in files1_rel:
        if(file_i in files2_rel_set and dr1[file_i] != dr2[file_i]):
            IF_AT_LEAST_ONE = True
            file1_i_abs = os.path.join(folder1_abs, file_i)
            file2_i_abs = os.path.join(folder2_abs, file_i)
            pout("*\t\t\t==========Difx changed==========")
            pout(f"\t\"{file1_i_abs}\" with hash=\"{dr1[file_i]}\"")
            pout(f"\t\"{file2_i_abs}\" with hash=\"{dr2[file_i]}\"\n")
    
    for file_i in files1_rel:
        if(file_i not in files2_rel_set):
            pout("*\t\t\t==========Difx new/removed==========")
            IF_AT_LEAST_ONE = True
            pout(f"\t\"{file_i}\"     IN \"{folder1_abs}\"")
            pout(f"\t\"{file_i}\" NOT IN \"{folder2_abs}\"")

    for file_i in files2_rel:
        if(file_i not in files1_rel_set):
            pout("*\t\t\t==========Difx new/removed==========")
            IF_AT_LEAST_ONE = True
            pout(f"\t\"{file_i}\" NOT IN \"{folder1_abs}\"")
            pout(f"\t\"{file_i}\"     IN \"{folder2_abs}\"")

    if(IF_AT_LEAST_ONE == False):
        pout("\tNo such files")

    if(len(err_out) != 0):
        pout(f"\n===============\nSome troubles happened:")
        for err_i in err_out:
            pout(f"\t{err_i}")
        pout(f"===============")

    pout("=============== Done! ===============")



# -*- coding: utf-8 -*-

import os
import sys
import argparse
import re
import subprocess
import time

from diwork_ways import *
from diwork_sup import *

# TODO: check how wort with sshpass

def process_src_dest_or_exit(folder1: str, folder2: str) -> "tuple(sorce, dest, user, host, path, if_src_remote)":
    req = f"(.+)@(.+):(.+)"
    re_res1 = re.match(req, folder1)
    re_res2 = re.match(req, folder2)
    if(   (re_res1 == None and re_res2 == None) or (re_res1 != None and re_res2 != None)   ):
        pout(f"Only one of \"{folder1}\" and \"{folder2}\" must be sshd (\"user@IP:path/to/dir\")")
        exit()
    
    if(re_res1 == None):
        if(is_folder(folder1) == False):
            pout(f"\"{folder1}\" must be folder. ")
            exit()
        else:
            re_res1 = os.path.abspath(folder1)
    else:
        pout(f"Will be connect to {re_res1[1]} on remote host {re_res1[2]} and use directory: {re_res1[3]}")
        res_user, res_host, res_path = re_res1[1], re_res1[2], re_res1[3]
        if_src = True

    if(re_res2 == None):
        if(is_folder(folder2) == False):
            pout(f"\"{folder2}\" must be folder. ")
            exit()
        else:
            re_res2 = os.path.abspath(folder2)
    else:
        pout(f"Will be connect to {re_res2[1]} on remote host {re_res2[2]} and use directory: {re_res2[3]}")
        res_user, res_host, res_path = re_res2[1], re_res2[2], re_res2[3]
        if_src = False

    return (re_res1, re_res2, res_user, res_host, res_path, if_src)
    

def run_remote_host_command(command: str, user: str, host: str, port: int, pswd:str=None) -> tuple:
    ''' return tuple of (stdout, stderr, returncode) '''
    if(pswd == None):
        com = f"ssh -p {port} {user}@{host} \"{command}\""
        DEBUG = True
    else:
        com = f"sshpass -p \"{pswd}\" ssh -p {port} {user}@{host} \"{command}\""
        com_out = f"sshpass -p *** ssh -p {port} {user}@{host} \"{command}\""
        DEBUG = False
        pout(f"> {com_out}")
    exe_res = exe(com, debug=DEBUG)
    #print(exe_res)
    #exit()
    return exe_res

def get_all_files_from_remote_host(user: str, host: str, port: int, path: str, pswd:str=None) -> list:
    com = f"find {path} -type f"
    #print(exe_res)
    exe_res = run_remote_host_command(com, user, host, port, pswd)
    if(exe_res[1] != ""):
        pout(f"=====ERROR=====\n{exe_res[1]}\n===============")
        exit()
    else:
        files = exe_res[0].split("\n")
    #print(files)
    return files[:len(files)-1] # last element is empty string

def do_copy(if_src_remote: bool, src: str, dest: str, port: int, pswd:str=None):
    if(pswd != None):
        com_prefix = f"sshpass -p \"{pswd}\" "
        com_prefix_out = f"sshpass -p *** "
        DEBUG = False
    else:
        com_prefix = ""
        DEBUG = True
    
    com = f"scp -P {port} \"{src}\" \"{dest}\""
    while(True):
        exe_res = exe(com_prefix + com, debug=DEBUG, std_out_fd=None, std_err_fd=subprocess.PIPE)
        #print(exe_res)
        if(exe_res[1] != ""):
            pout(f"Error: {exe_res[1]}\n Trying again...\n")
            time.sleep(3)
        else:
            break

    if(DEBUG == False):
        pout(f"> {com_prefix_out + com}")

def delete_all_if_dir_not_empty_ssh(dir_path: str, user: str, host: str, port: int = 22, pswd:str=None):
    exe_res = run_remote_host_command(f"ls \"{dir_path}\"", user, host, port, pswd)
    if(exe_res[1] != ""):
        pout(f"=====ERROR=====\n{exe_res[1]}\n===============")
        exit()
    if(exe_res[0] != ""):
        ssh_folder_name = f"{user}@{host}:{dir_path}"
        #rm_command = f"rm -rf {path}"
        #rm_command = f"find {path} -type f -delete"
        rm_command = f"find \"{dir_path}\" ! -wholename \"{dir_path}\" -delete"

        pout(f"Folder \"{ssh_folder_name}\" is not empty. ")
        pout(f"===============\n\t All files in \"{ssh_folder_name}\" will be removed before continue. \n\tOn server \"{host}\" command \"\"\"\n> {rm_command}\n\"\"\" will be executed. \n===============")
        pout("Continue? Type \"yes\" in capital letter to continue or \"no\" to exit. \n> ", endl=False)
        
        while(True):
            user_in = input().strip()
            if(user_in != "YES"):
                pout("Type \"YES\" in capital letter to remove all content of directory \"{dir_path}\" or press CTRL+C or type \"no\" to exit. ")
            if(user_in.lower() == "no"):
               pout("Exitting...")
               exit()
            if(user_in == "YES"):
                break
            pout("> ", endl=False)

        exe_res = run_remote_host_command(rm_command, user, host, port, pswd)
        if(exe_res[1] != ""):
            pout(f"=====ERROR=====\n{exe_res[1]}\n===============")
            exit()
        
        exe_res = run_remote_host_command(f"ls \"{dir_path}\"", user, host, port, pswd)
        if(exe_res[1] != ""):
            pout(f"=====ERROR=====\n{exe_res[1]}\n===============")
            exit()
        if(exe_res[0] == ""):
            pout(f"All files from folder \"{ssh_folder_name}\" removed. This folder is empty now.")
        else:
            pout(f"Cannot clean folder \"{ssh_folder_name}\"! Exiting ")
            exit()



def main_sshclone(args: list):

    parser = argparse.ArgumentParser(prog = "diwork sshclone",
        description="Clone directory by scp. Set up ssh-agent. And run \"ssh-add\" or use flag --askpass (sshpass must be installed). ")
    parser.add_argument("source", type=str, nargs=1,
                       help="Path to directory or user@IP/DOMEN:path/to/dir")
    parser.add_argument("destination", type=str, nargs=1,
                       help="Path to directory or user@IP/DOMEN:path/to/dir")
    parser.add_argument("--askpass", default=False, action='store_true',
                       help="If ssh do not use key. Then it will ask password. Package \"sshpass\" must be installed. ")
    parser.add_argument("--port", type=int, default=22, 
                       help="port of sshd. ")
    parser = common_init_parser(parser)
    args = parser.parse_args(args)
    common_init_parse(args)

    folder1 = args.source[0]
    folder2 = args.destination[0]
    IF_ASKPASS = args.askpass
    PORT = args.port


    folder1, folder2, user, host, path, if_src_remote = process_src_dest_or_exit(folder1, folder2)

    if(IF_ASKPASS == True):
        from getpass import getpass
        while(True):
            PSWD1 = getpass(f"Input password of {user}: ")
            PSWD2 = getpass(f"Input again password of {user}: ")
            if(PSWD1 != PSWD2):
                pout("Passwords do not match. Try again.")
            else:
                break
        PSWD = PSWD1
        
    else:
        PSWD = None

    exe_res = run_remote_host_command(f"[ -d \"{path}\" ] && echo -n 1", user, host, PORT, PSWD)
    if(exe_res[1] != ""):
        pout(f"=====ERROR=====\n{exe_res[1]}\n===============")
        exit()
    if(exe_res[0] != "1"):
        pout(f"Directory \"{user}@{host}:{path}\" does not exists. ")
        exit()

    if(if_src_remote == True):
        cpfiles = get_all_files_from_remote_host(user, host, PORT, path, PSWD)
    else:
        cpfiles = get_files_list(folder1)

    if(if_src_remote == False):
        delete_all_if_dir_not_empty_ssh(path, user, host, PORT, PSWD)
        dirs = get_dirs_needed_for_files(cpfiles)
        for dir_i in dirs:
            dir_i_rel = rel_path(dir_i, folder1)
            dir_i_2 = os.path.join(path, dir_i_rel)
            com = f"mkdir -p \"{dir_i_2}\""
            exe_res = run_remote_host_command(com, user, host, PORT, PSWD)
            if(exe_res[1] != ""):
                pout(f"=====ERROR=====\n{exe_res[1]}\n===============")
                exit()


    if(if_src_remote == True):
        #pout(f"Folder \"{folder2}\" is not empty. ")
        #exit()
        delete_all_if_dir_not_empty(folder2)

        dirs = get_dirs_needed_for_files(cpfiles)
        for dir_i in dirs:
            dir_i_rel = rel_path(dir_i, path)
            dir_i_2 = os.path.join(folder2, dir_i_rel)
            exe_out = exe(f"mkdir -p \"{dir_i_2}\"")
            if(exe_out[1] != ""):
                pout(f"ERROR: {exe_out[1]}")
                exit()

    if(if_src_remote == True):
        gi, N = 1, len(cpfiles)
        for file_i in cpfiles:
            file_i_rel = rel_path(file_i, path)
            src_file, dest_file = f"{user}@{host}:{file_i}",  os.path.join(folder2, file_i_rel)
            pout(f"({gi}/{N}) Coping file \"{src_file}\" to \"{dest_file}\"...")
            do_copy(if_src_remote, src_file,dest_file, PORT, PSWD)
            gi-=-1
    else:
        gi, N = 1, len(cpfiles)
        for file_i in cpfiles:
            file_i_rel = rel_path(file_i, folder1)
            dest_file = f"{user}@{host}:{os.path.join(path, file_i_rel)}"
            pout(f"({gi}/{N}) Coping file \"{file_i}\" to \"{dest_file}\"...")
            do_copy(if_src_remote, file_i, dest_file, PORT, PSWD)
            gi-=-1





    pout("\n=============== Done! ===============")



# -*- coding: utf-8 -*-

from .diwork_sup import *

from .diwork_parse import common_init_parser, common_init_parse

__all__ = [
    "Global", "get_files_list", "getDirsList", "is_folder", "is_file", "is_exists", "is_folder_empty", "rel_path", "rm_folder_content",
    "check_files_exists_or_exit", "pout", "write2File_str", "mkdir", "get_link_unwinding", "get_nice_size", "get_time_str", "get_time_file",
    "copy_file", "get_hash_file", "get_hash_str", "get_hash_of_hashes",
    "exclude_files", "delete_all_if_dir_not_empty", "exe",
    "common_init_parser", "common_init_parse",
    "ThreadWithReturnValue"
]


# -*- coding: utf-8 -*-

import os
import sys

from diwork_ways import pout, Global

from diwork_mains import *

from . import __version__

VERSION = __version__

def main():
    MODULES = "{help, hash, clone, diffclone, sshclone, diff, difx, repeats, exec, archive}"
    SyntaxError_str = f"Syntax error. Expected: \"> python folder_work.py {MODULES} ...\""
    argc = len(sys.argv)
    Global.version = VERSION
    if(argc < 2):
        pout(SyntaxError_str)
        exit()
    else:
        sub_modul_name = sys.argv[1]
        if(sub_modul_name == "hash"):
            main_hash(sys.argv[2:])
        elif(sub_modul_name == "clone"):
            main_clone(sys.argv[2:])
        elif(sub_modul_name == "diffclone"):
            main_diffclone(sys.argv[2:])
        elif(sub_modul_name == "sshclone"):
            main_sshclone(sys.argv[2:])
        elif(sub_modul_name == "diff"):
            main_diff(sys.argv[2:])
        elif(sub_modul_name == "repeats"):
            main_repeats(sys.argv[2:])
        elif(sub_modul_name == "difx"):
            main_difx(sys.argv[2:])
        elif(sub_modul_name == "exec"):
            main_exec(sys.argv[2:])
        elif(sub_modul_name == "archive"):
            main_archive(sys.argv[2:])
        elif(sub_modul_name == "help"):
            main_help(sys.argv[2:], MODULES)
        else:
            pout(SyntaxError_str)
            exit()




# -*- coding: utf-8 -*-

import os
import sys
import subprocess
import hashlib
import datetime
import shutil
from threading import Thread


class Global():
    version = None
    outfile = None
    symlink_mode = None


class ThreadWithReturnValue(Thread):

    def __init__(self, index: int, group=None, target=None, name=None,
                 args=(), kwargs={}, Verbose=None):
        Thread.__init__(self, group, target, name, args, kwargs)
        self._index = index
        self._return = None

    def run(self):
        if self._target is not None:
            self._return = self._target(*self._args,
                                        **self._kwargs)

    def get_index(self):
        return self._index

    def join(self, *args):
        Thread.join(self, *args)
        return self._return


def get_files_list(dirPath: str) -> list:
    return [os.path.join(path, name) for path, subdirs, files in os.walk(dirPath) for name in files]


def getDirsList(dirPath: str) -> list:
    dirs = [os.path.join(path, name) for path, subdirs, files in os.walk(dirPath) for name in subdirs]
    return list(set(dirs))


# return False, if folder_path not exists or folder_path is not folder
def is_folder(folder_path: str) -> bool:
    return os.path.isdir(folder_path)


# return False, if file_path not exists or file_path is not file
def is_file(file_path: str) -> bool:
    return os.path.isfile(file_path)


# return False, if file_path not exists
def is_exists(file_path: str) -> bool:
    return os.path.exists(file_path)


def check_files_exists_or_exit(files: list) -> None:
    F = False
    for file_i in files:
        if(is_exists(file_i) == False):
            pout(f"File \"{file_i}\" does not exists. ")
            F = True
    if(F == True):
        exit()


def is_folder_empty(folder_path: str) -> bool:
    if(len(os.listdir(folder_path)) == 0):
        return True
    else:
        return False


def rel_path(file_path: str, folder_path: str) -> str:
    return os.path.relpath(file_path, folder_path)


def rm_folder_content(folder_path: str, root_dir_too: bool = False, does_not_exists_is_ok = False):
    """Удаляет всё содержимое папки. Саму папку не трогает, если root_dir_too == False"""
    if(does_not_exists_is_ok == True and is_folder(folder_path) == False):
        return
    for root, dirs, files in os.walk(folder_path, topdown=False):
        for file_i in files:
            os.remove(os.path.join(root, file_i))
        for dir in dirs:
            os.rmdir(os.path.join(root, dir))
    if(root_dir_too == True):
        os.rmdir(folder_path)


def pout(msg: str, endl = True):
    if(endl == False):
        pout_low(msg)
    else:
        pout_low(msg + "\n")


def pout_low(msg: str):
    print(msg, end="")
    if(Global.outfile != None):
        with open(Global.outfile, "a", encoding="utf-8") as fd:
            fd.write(msg)
            fd.flush()


def write2File_str(fileName : str, s : str) -> None:
    with open(fileName, 'w', encoding="utf-8") as temp:
        temp.write(s)
        temp.flush()


def mkdir(path: str, p: bool = True):
    os.makedirs(path, exist_ok=True)


def get_link_unwinding(link_path: str) -> str or None:
    """Вернёт конечный файл, на который (рекурсивно) ссылаются сылки. """
    if(os.path.exists(link_path) == False):
        return None
    elif(os.path.islink(link_path) == False):
        return link_path
    else:
        linkto = os.readlink(link_path)
        if(os.path.islink(linkto) == False):
            return linkto
        else:
            return get_link_unwinding(linkto)


def get_nice_size(size_bytes: int) -> str:
    if(size_bytes < 1024):
        return f"{size_bytes} B"
    elif(size_bytes < 1024*1024):
        return f"{size_bytes // 1024} KB"
    elif(size_bytes < 1024*1024*1024):
        return f"{size_bytes // (1024*1024)} MB"
    elif(size_bytes < 1024*1024*1024*1024):
        return f"{size_bytes // (1024*1024*1024)} GB"
    else:
        return f"{size_bytes // (1024*1024*1024*1024)} TB"


def get_time_str() -> str:
    # time_str = datetime.datetime.now().strftime("[%y.%m.%d %H:%M:%S.%f]")
    time_str = datetime.datetime.now().strftime("%y.%m.%d %H:%M:%S")
    return time_str


def get_time_file(file_path: str) -> str:
    dt_m = datetime.datetime.fromtimestamp(os.path.getmtime(file_path))
    return dt_m.strftime("%d.%m.%Y_%H-%M-%S")


def copy_file(src: str, dest: str) -> bool or None:
    if os.path.islink(src):
        if Global.symlink_mode == 0:
            return None
        elif Global.symlink_mode == 1:
            linkto = os.readlink(src)
            os.symlink(linkto, dest)
            return True
        elif Global.symlink_mode == 2:
            file_path = get_link_unwinding(src)
            if file_path is None:
                return False
        else:
            raise Exception(f"Failed successfully: copy_file with {src}->{dest}")

    shutil.copy(src, dest)
    return True


def get_hash_file(file_path: str) -> str or None:
    if os.path.islink(file_path):
        if Global.symlink_mode == 0:
            return None
        elif Global.symlink_mode == 1:
            return get_hash_str(os.readlink(file_path))
        elif Global.symlink_mode == 2:
            file_path = get_link_unwinding(file_path)
            if file_path is None:
                return None
        else:
            raise Exception(f"Failed successfully: get_hash_file with {file_path}")

    buff_BLOCKSIZE = 65536  # 64 kB
    sha = hashlib.sha256()
    with open(file_path, "rb") as temp:
        file_buffer = temp.read(buff_BLOCKSIZE)
        while len(file_buffer) > 0:
            sha.update(file_buffer)
            file_buffer = temp.read(buff_BLOCKSIZE)
    return sha.hexdigest()


def get_hash_str(s: str):
    return hashlib.sha256( s.encode("utf-8") ).hexdigest()


def get_hash_of_hashes(hashes: list) -> str:

    # IF CHANGE, then change make_archive_one_folder in diwork_archive.py (its legacy_version)

    # from io import StringIO
    # o = StringIO()
    # for hash_i in hashes:
    #     o.write(hash_i)
    # hash_files = get_hash_str(o.getvalue())
    # return hash_files
    hash_files = ""
    li = 0
    for hash_i in hashes:
        hash_files += hash_i
        li-=-1
        if(li == 30):
            hash_files = get_hash_str(hash_files)
            li = 0
    hash_files = get_hash_str(hash_files)
    return hash_files


def exclude_files(src_files: list, exclude_files: list) -> list:
    """
    Убрать из src те файлы, которые есть в exclude_files
    exclude_files - это список файлов или директорий
    Пути к файлам либо все абсолютные, либо все относительные
    """
    if(exclude_files == None):
        return src_files
    res = []

    for file_i in src_files:
        if file_i not in exclude_files:
            F = True
            for ex_file_i in exclude_files:
                if ex_file_i in file_i:
                    F = False
            if(F == True):
                res.append(file_i) 
    return res


def delete_all_if_dir_not_empty(dir_path: str):
    if(is_folder_empty(dir_path) == False):
        pout(f"Folder \"{dir_path}\" is not empty. ")
        pout(f"===============\n\t All files in \"{dir_path}\" will be removed before continue. \n===============")
        pout("Continue? Type \"yes\" in capital letter to continue or \"no\" to exit. \n> ", endl=False)
        
        while(True):
            user_in = input().strip()
            if(user_in != "YES"):
                pout("Type \"YES\" in capital letter to remove all content of directory \"{dir_path}\" or press CTRL+C or type \"no\" to exit. ")
            if(user_in.lower() == "no"):
               pout("Exitting...")
               exit()
            if(user_in == "YES"):
                break
            pout("> ", endl=False)
        rm_folder_content(dir_path)
        if(is_folder_empty(dir_path) == True):
            pout(f"All files from folder \"{dir_path}\" removed. This folder is empty now.")
        else:
            pout(f"Cannot clean folder \"{dir_path}\"! Exiting ")
            exit()


def get_dirs_needed_for_files(files: list) -> list:
    dirs = set()
    for file_i in files:
        dir_i = os.path.dirname(file_i)
        dirs.add(dir_i)
    dirs = sorted(list(dirs))
    return dirs


def exe_lowout(command: str, debug: bool = True, std_out_pipe: bool = False, std_err_pipe: bool = False) -> tuple:
    '''
    Аргумент command - команда для выполнения в терминале. Например: "ls -lai ."
    Возвращает кортеж, где элементы:
        0 - строка stdout or None if std_out_pipe == False
        1 - строка stderr or None if std_err_pipe == False
        2 - returncode
    '''
    if(debug):
        pout(f"> {command}")
    
    if(std_out_pipe == True):
        if(std_err_pipe == True):
            process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            # https://stackoverflow.com/questions/1180606/using-subprocess-popen-for-process-with-large-output
            out = process.stdout.read().decode("utf-8")
            err = process.stderr.read().decode("utf-8")
        else:
            process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
            out = process.stdout.read().decode("utf-8")
            err = None
    else:
        if(std_err_pipe == True):
            process = subprocess.Popen(command, shell=True, stderr=subprocess.PIPE)
            out = None
            err = process.stderr.read().decode("utf-8")
        else:
            process = subprocess.Popen(command, shell=True)
            out = None
            err = None
    errcode = process.returncode
    return (out, err, errcode)


def exe(command: str, debug: bool = True, std_out_fd = subprocess.PIPE, std_err_fd = subprocess.PIPE, stdin_msg: str = None) -> tuple:
    '''
    Аргумент command - команда для выполнения в терминале. Например: "ls -lai ."
    if(std_out_fd or std_err_fd) == subprocess.DEVNULL   |=>    No output enywhere
    if(std_out_fd or std_err_fd) == subprocess.PIPE      |=>    All output to return
    if(std_out_fd or std_err_fd) == open(path, "w")      |=>    All output to file path
    Возвращает кортеж, где элементы:
        0 - строка stdout
        1 - строка stderr
        2 - returncode
    '''
    _ENCODING = "utf-8"

    if(debug):
        #pout(f"> " + " ".join(command))
        if(stdin_msg != None):
            pout(f"> {command}, with stdin=\"{stdin_msg}\"")
        else:
            pout(f"> {command}")

    #proc = subprocess.run(command, shell=True, capture_output=True, input=stdin_msg.encode("utf-8"))
    if(stdin_msg == None):
        proc = subprocess.run(command, shell=True, stdout=std_out_fd, stderr=std_err_fd)
    else:
        proc = subprocess.run(command, shell=True, stdout=std_out_fd, stderr=std_err_fd, input=stdin_msg.encode("utf-8"))
    
    #return (proc.stdout.decode("utf-8"), proc.stderr.decode("utf-8"))

    res_stdout = proc.stdout.decode("utf-8") if proc.stdout != None else None
    res_errout = proc.stderr.decode("utf-8") if proc.stderr != None else None
    return (res_stdout, res_errout, proc.returncode)



# -*- coding: utf-8 -*-

import os
import sys
import argparse 

from diwork_ways import *

def main_help(args: list, modules: str):
    text = f"""
Available modules: {modules}.
To choose a module, run the command:
> diwork {'{module_name}'} ...

For example:
> diwork hash ...

If the syntax is incorrect, it will show the help and description of the selected module.

For example. To show the help and description of module clone, run the command:
> diwork clone

Some modules require the help flag to be specified explicitly. For example:
> diwork hash --help
"""
    pout(text)



# -*- coding: utf-8 -*-

# For pycharm: Setting -> Project -> Project Structure ->
#                           Mark as source all folders contains src

from .diwork_hash import main_hash
from .diwork_clone import main_clone
from .diwork_diff import main_diff
from .diwork_difx import main_difx
from .diwork_repeats import main_repeats
from .diwork_exec import main_exec
from .diwork_help import main_help
from .diwork_sshclone import main_sshclone
from .diwork_archive import main_archive
from .diwork_diffclone import main_diffclone

__all__ = [
    "main_help",
    "main_hash",
    "main_clone",
    "main_diff",
    "main_difx",
    "main_repeats",
    "main_exec",
    "main_sshclone",
    "main_archive",
    "main_diffclone"
]



# -*- coding: utf-8 -*-

import os
import sys
import argparse 

from diwork_ways import *


def diff_new_files(dr1, dr2) -> list:
    '''Do not need hash'''
    files_rel_old, files_rel_new = set(dr1.keys()), set(dr2.keys())
    res = []
    for file_i in files_rel_new:
        if(file_i not in files_rel_old):
            res.append(file_i)
    return res

def diff_changes_files(dr1, dr2) -> list:
    files_rel_old, files_rel_new = set(dr1.keys()), set(dr2.keys())
    res = []
    for file_i in files_rel_new:
        if(file_i in files_rel_old):
            if(dr1[file_i] != dr2[file_i]):
                res.append(file_i)
    return res

def diff_removed_files(dr1, dr2) -> list:
    '''Do not need hash'''
    files_rel_old, files_rel_new = set(dr1.keys()), set(dr2.keys())
    res = []
    for file_i in files_rel_old:
        if(file_i not in files_rel_new):
            res.append(file_i)
    return res

def diff_moved_files(dr1, dr2, old_folder_path, new_folder_path) -> list:
    files_rel_old, files_rel_new = set(dr1.keys()), set(dr2.keys())
    res = []
    for file_i_old in files_rel_old:
        S = f"File \"[{old_folder_path}] {file_i_old}\" has been moved (renamed) to: \n"
        S += f"\t\t{dr1[file_i_old]}\n"
        IF_AT_LEAST_ONE = False
        for file_i_new in files_rel_new:
            if(dr1[file_i_old] == dr2[file_i_new]):
                if(file_i_old != file_i_new):
                    if(file_i_new not in files_rel_old):
                        IF_AT_LEAST_ONE = True
                        S += f"\t[{new_folder_path}] {file_i_new}"
        if(IF_AT_LEAST_ONE == True):
            res.append(S)
    return res

def diff_identical_files(d1, d2, old_folder_path, new_folder_path) -> list:
    len_diff = len(old_folder_path) - len(new_folder_path)
    old_folder_prefix, new_folder_prefix = "", ""
    if(len_diff > 0):
        new_folder_prefix = " "*len_diff
    else:
        old_folder_prefix = " "*(-len_diff)
    files_abs_old, files_abs_new = set(d1.keys()), set(d2.keys())
    hashes1, hashes2 = list(d1.values()), list(d2.values())
    hashes = set(hashes1 + hashes2)
    res = []
    for hash_i in hashes:
        item_fits = 0
        S = f"* Hash \"{hash_i}\" have files: \n"
        for file_i in files_abs_new:
            if(d2[file_i] == hash_i):
                item_fits += 1
                S += f"\t- {new_folder_prefix}{file_i}\n"
        for file_i in files_abs_old:
            if(d1[file_i] == hash_i):
                item_fits += 1
                S += f"\t- {old_folder_prefix}{file_i}\n"
        if(item_fits >= 2):
            res.append(S)
    return res

def main_diff(args: list):
    argc = len(args)
    mode_explain_str = "\t- n: show New files (does not need the hash)\n\t- c: show Changed files\n\t- r: show Removed files (does not need the hash)\n\t- m: show Moved (renamed) files\n\t- i: show Identical files"
    if(argc != 3):
        pout("This module will show all changes in directories.\n")
        pout("Syntax error. Expected: \"python folder_work.py diff {n|c|r|m|i} {folder_old} {folder_new}\", where: ")
        pout(mode_explain_str)
        exit()
    mode = args[0]
    folder_old = args[1]
    folder_new = args[2]
    err_out = []
    folder_old_abs = os.path.abspath(folder_old)
    folder_new_abs = os.path.abspath(folder_new)
    if(is_folder(folder_old_abs) == False):
        pout(f"\"{folder_old_abs}\" is not folder. ")
        exit()
    if(is_folder(folder_new_abs) == False):
        pout(f"\"{folder_new_abs}\" is not folder. ")
        exit()
    for letter_i in mode:
        if(letter_i not in "ncrmi"):
            pout(f"Cannot understand \"{letter_i}\" in \"{mode}\". Expected: n, c, r, m, i or their combination: ")
            pout(mode_explain_str)
            exit()
        if(mode.count(letter_i) > 1):
            pout(f"\"{letter_i}\" cannot repeat. ")
            exit()
    if(len(mode) == 2 and mode[0] in "nr" and mode[1] in "nr"):
        HASH_NOT_NEEDED = True
    else:
        HASH_NOT_NEEDED = False
    d1, d2, dr1, dr2 = {}, {}, {}, {}
    files_abs_old = sorted(get_files_list(folder_old_abs))
    files_abs_new = sorted(get_files_list(folder_new_abs))
    gi, N = 0, len(files_abs_old)+len(files_abs_new)
    for file_i in files_abs_old:
        gi+=1
        if(is_file(file_i) == False):
            err_out.append(f"\"{file_i}\" is not file or does not exists, it will be skipped. ")
            continue
        if(HASH_NOT_NEEDED == True):
            file_i_hash = ""
        else:
            file_i_hash = get_hash_file(file_i)
            pout(f"({gi}/{N}) Calculated hash of \"{file_i}\": {file_i_hash}")
        d1[file_i] = file_i_hash
        file_i_rel = rel_path(file_i, folder_old_abs)
        dr1[file_i_rel] = file_i_hash
    for file_i in files_abs_new:
        gi+=1
        if(is_file(file_i) == False):
            err_out.append(f"\"{file_i}\" is not file or does not exists, it will be skipped. ")
            continue
        if(HASH_NOT_NEEDED == True):
            file_i_hash = ""
        else:
            file_i_hash = get_hash_file(file_i)
            pout(f"({gi}/{N}) Calculated hash of \"{file_i}\": {file_i_hash}")
        d2[file_i] = file_i_hash
        file_i_rel = rel_path(file_i, folder_new_abs)
        dr2[file_i_rel] = file_i_hash

    pout("\nHashes calculated. Calculating differences. \n")

    for mode_i in mode:
        if(mode_i == "n"):
            new_files_rel = diff_new_files(dr1, dr2)
            pout("\n====================\n* New files: ")
            if(len(new_files_rel) > 0):
                for file_i in new_files_rel:
                    pout(f"\t{os.path.join(folder_new_abs, file_i)}")
            else:
                pout("\tNo such files. ")
        elif(mode_i == "c"):
            changed_files_rel = diff_changes_files(dr1, dr2)
            pout("\n====================\n* Changed files: ")
            if(len(changed_files_rel) > 0):
                for file_i in changed_files_rel:
                    pout(f"\t{file_i}")
            else:
                pout("\tNo such files. ")
        elif(mode_i == "r"):
            removed_files_rel = diff_removed_files(dr1, dr2)
            pout("\n====================\n* Removed files: ")
            if(len(removed_files_rel) > 0):
                for file_i in removed_files_rel:
                    pout(f"\t{os.path.join(folder_old_abs, file_i)}")
            else:
                pout("\tNo such files. ")
        elif(mode_i == "m"):
            moved_files_rel = diff_moved_files(dr1, dr2, folder_old_abs, folder_new_abs)
            pout("\n====================\n* Moved (renamed) files: ")
            if(len(moved_files_rel) > 0):
                for file_i in moved_files_rel:
                    pout(f"{file_i}")
            else:
                pout("\tNo such files. ")
        elif(mode_i == "i"):
            identical_files = diff_identical_files(d1, d2, folder_old_abs, folder_new_abs)
            pout("\n====================\n* Identical files: ")
            if(len(identical_files) > 0):
                for file_i in identical_files:
                    pout(f"{file_i}")
            else:
                pout("\tNo such files. ")
        else:
            pout(f"Failed successfully (\"{mode_i}\").")

    if(len(err_out) != 0):
        pout(f"\n===============\nSome troubles happened:")
        for err_i in err_out:
            pout(f"\t{err_i}")
        pout(f"===============")

    pout("=============== Done! ===============")



# -*- coding: utf-8 -*-

import os
import sys

__WORK_DIR = os.path.dirname(__file__)
sys.path.insert(0, __WORK_DIR)

sys.path.insert(0, os.path.abspath(f"{__WORK_DIR}/diwork_ways"))
sys.path.insert(0, os.path.abspath(f"{__WORK_DIR}/diwork_mains"))

from .__version__ import __version__
from .main import main

__all__ = ["__version__", main]



# -*- coding: utf-8 -*-

import os
import sys
import argparse

from . import Global

def common_init_parser(parser: "ArgumentParser") -> "ArgumentParser":
    # https://docs.python.org/3/library/argparse.html
    # https://stackoverflow.com/questions/20063/whats-the-best-way-to-parse-command-line-arguments

    parser.add_argument("--dublicate_out_to_file", type=str, default=None, required=False,
                       help="Duplicate program output to file")

    parser.add_argument("--version", action="version", version=f"diwork {Global.version}",
                       help="Check version of diwork")

    parser.add_argument("--symlink_mode", type=int, choices=[0, 1, 2], default=2, required=False,
                       help="What to do with links: 0 - ignor, 1 - consider link content, 2 - use the file where the link refers to. Default 2.")
    return parser

def common_init_parse(args: "ArgumentParser.parse_args") -> None:
    Global.outfile = args.dublicate_out_to_file
    Global.symlink_mode = args.symlink_mode



# -*- coding: utf-8 -*-

import os
import sys
import argparse
import shutil
import threading

from diwork_ways import *

from tqdm import tqdm


def calc_hash_dict_of_dir(dir_path: str) -> dict:
    d = {}
    dir_path = os.path.abspath(dir_path)
    files = get_files_list(dir_path)
    files = sorted(files)
    for file_i in tqdm(files):
        file_i_rel = os.path.relpath(file_i, dir_path)
        d[file_i_rel] = get_hash_file(file_i)
    return d


def main_diffclone(args: list):
    platform = sys.platform
    parser = argparse.ArgumentParser(prog = "diwork diffclone",
        description="This module will clone all the contents of {folder_src} to {folder_dest}, but only files, which does not already contains in {folder_dest}. "
                    "Old files will be renamed if needed.")
    parser.add_argument("folder_src", type=str, nargs=1,
                       help="Path to source directory")
    parser.add_argument("folder_dest", type=str, nargs=1,
                       help="Path to destination directory")
    parser.add_argument("--no_check_hash", default=False, action='store_true',
                       help="Check hashes of files from {folder_src} and {folder_dest}")
    parser.add_argument("--skip_copy", default=False, action='store_true',
                       help="Skip copy phase. Only hash ckeck.")

    parser = common_init_parser(parser)
    args = parser.parse_args(args)
    common_init_parse(args)

    folder1 = args.folder_src[0]
    folder2 = args.folder_dest[0]
    err_out = []
    check_hash = not args.no_check_hash
    skip_copy = args.skip_copy
    folder1_abs = os.path.abspath(folder1)
    folder2_abs = os.path.abspath(folder2)
    if not is_folder(folder1_abs):
        pout(f"\"{folder1_abs}\" is not folder. ")
        exit()
    if not is_folder(folder2_abs):
        pout(f"\"{folder2_abs}\" is not folder. ")
        exit()

    if folder1_abs in folder2_abs:
        pout(f"Directory \"{folder1_abs}\" contains directory \"{folder2_abs}\". ")
        input("Enter to continue...")
    if folder2_abs in folder1_abs:
        pout(f"Directory \"{folder2_abs}\" contains directory \"{folder1_abs}\". ")
        input("Enter to continue...")

    pout(f"Calculating hash tree of directory \"{folder1_abs}\"...")
    # d1 = calc_hash_dict_of_dir(folder1_abs)
    cp_count = 0

    pout("Diffclonning...")
    # 3 варианта:
    # 1. Такого файла не существует в folder2, тогда просто копируем
    # 2. Такой файл существует в folder2 и hash совпадает, тогда ничего не делаем
    # 3. Такой файл существует в folder2 и hash не совпадает, тогда старый переименовываем, а новый копируем

    dirs_abs_1 = getDirsList(folder1_abs)
    dirs_abs_1 = sorted(dirs_abs_1)
    files1, files2 = get_files_list(folder1_abs), get_files_list(folder2_abs)
    
    if not skip_copy:
        for dir_i_1 in dirs_abs_1:
            dir_i_rel = rel_path(dir_i_1, folder1_abs)
            dir_i_2 = os.path.join(folder2_abs, dir_i_rel)
            if not os.path.exists(dir_i_2):
                mkdir(dir_i_2, p=True)

        for file1_i in tqdm(files1):
            try:
                if os.path.islink(file1_i):
                    if Global.symlink_mode == 0:
                        continue

                file1_i_rel = str(os.path.relpath(file1_i, folder1_abs))
                file2_i = os.path.join(folder2_abs, file1_i_rel)
                if not os.path.isfile(file2_i):  # case 1
                    if copy_file(file1_i, file2_i) in [False, None]:
                        err_out.append(f"Cannot copy \"{file1_i}\" to \"{file2_i}\". ")
                    else:
                        cp_count += 1
                else:
                    x1 = ThreadWithReturnValue(1, target=get_hash_file, args=(file1_i,))
                    x2 = ThreadWithReturnValue(2, target=get_hash_file, args=(file2_i,))
                    x1.start(), x2.start()
                    file1_i_hash, file2_i_hash = x1.join(), x2.join()
                    if file1_i_hash == file2_i_hash:  # case 2
                        pass  # ничего не делаем
                    else:  # case 3
                        file2_i_time = get_time_file(file2_i)
                        os.rename(file2_i, f"{file2_i}---{file2_i_time}.bak")
                        if copy_file(file1_i, file2_i) in [False, None]:
                            err_out.append(f"Cannot copy \"{file1_i}\" to \"{file2_i}\". ")
                        else:
                            cp_count += 1
            except Exception as e:
                pout(str(e))
                err_out.append(f"\"{file1_i}\"")

        if platform != "win32":
            os.sync()
        if len(err_out) != 0:
            pout(f"\n===============\nSome troubles happened:")
            for err_i in err_out:
                pout(f"\t{err_i}")
            pout(f"===============")

        pout(f"Coping done! {cp_count} copy operation total. ")

    IF_OK = True
    if check_hash:
        err_out = []
        pout("Checking...")
        for file1_i in tqdm(files1):
            try:
                if os.path.islink(file1_i):
                    if Global.symlink_mode == 0:
                        continue
                file1_i_rel = str(os.path.relpath(file1_i, folder1_abs))
                file2_i = os.path.join(folder2_abs, file1_i_rel)
                x1 = ThreadWithReturnValue(1, target=get_hash_file, args=(file1_i,))
                x2 = ThreadWithReturnValue(2, target=get_hash_file, args=(file2_i,))
                x1.start(), x2.start()
                file1_i_hash, file2_i_hash = x1.join(), x2.join()
                if file1_i_hash != file2_i_hash:
                    buffS = (
                        f"HASHES OF FILES DOES NOT MATCH: "
                        f"\"{d1[file1_i_rel]}\" of \"{file1_i}\" and "
                        f"\"{file2_i_hash}\" of \"{file2_i}\""
                             )
                    pout(buffS)
                    err_out.append(buffS)
                    IF_OK = False
            except Exception as e:
                IF_OK = False
                pout(str(e))
                err_out.append(f"\"{file1_i}\"")
    if len(err_out) != 0:
        pout(f"\n===============\nSome troubles happened:")
        for err_i in err_out:
            pout(f"\t{err_i}")
        pout(f"===============")
    if not IF_OK:
        pout("Hashes does NOT match, diffclone failed!!!")
    else:
        pout("Hashes match! All is OK")
    pout("=============== Done! ===============")



# coding: utf-8

from setuptools import setup, find_packages
from distutils.util import convert_path

# https://setuptools.pypa.io/en/latest/userguide/quickstart.html
# pip install --upgrade pip
# pip install --upgrade build
# pip install --upgrade setuptools
# python3 -m pip install --upgrade twine
# python3 -m build
# python3 -m twine upload --repository testpypi dist/*
# pip3 install --no-deps --index-url https://test.pypi.org/simple {package_name}

main_ns = {}
ver_path = convert_path("diwork/__version__.py")
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

setup(
    name="diwork",
    version=main_ns["__version__"],
    install_requires=[
        "argparse",
        "tqdm",
    ],
    packages=find_packages(
        # All keyword arguments below are optional:
        where='.',  # '.' by default or "src"
    ),
    entry_points={
        "console_scripts": [
                "diwork = diwork.main:main",
            ]
    },
    include_package_data=False,
)



# -*- coding: utf-8 -*-

import os
import sys
import argparse

import json
import zipfile

from diwork_ways import *

def get_folders_one_deep(dir_path: str) -> list:
    # return [x[0] for x in os.walk(dir_path)]
    return [ f.path for f in os.scandir(dir_path) if f.is_dir() ]



def make_archive_one_folder(folder_path: str, path_to_zip: str, cur_iter: str = None) -> "(errors: list,)":
    """
    json:
    hashes: {file: hashs},
    sizes: {file: sizes},
    folder_size: int,
    folder_hash: hash,
    date: str(DD.MM.YY HH:MM:SS),
    legacy_version: int
    hash_mode: int
    """
    legacy_version = 1
    d = {"hashes":{}, "sizes": {}}

    errors = []
    folder_path_name = os.path.basename(folder_path)
    # folder_path_name = "content"
    files = sorted(get_files_list(folder_path))
    files = [os.path.abspath(file_i) for file_i in files]
    files_rel = [rel_path(file_i, folder_path) for file_i in files]

    with zipfile.ZipFile(path_to_zip, mode="w") as zfd:
        gi, gN = 0, len(files)
        for file_i, file_rel_i in zip(files, files_rel):
            gi-=-1
            hash_i = get_hash_file(file_i)
            size_i = os.path.getsize(file_i)
            d["hashes"][file_i] = hash_i
            d["sizes"][file_i] = size_i
            if(cur_iter != None):
                pout(f"{cur_iter} ({gi}/{gN}) Hash \"{hash_i}\" have file \"{file_i}\" ({get_nice_size(size_i)}). ")
            else:
                pout(f"({gi}/{gN}) Hash \"{hash_i}\" have file \"{file_i}\" ({get_nice_size(size_i)}). ")
            zfd.write(file_i, f"{folder_path_name}/{file_rel_i}")
        d["folder_hash"] = hash_files = get_hash_of_hashes(sorted(d["hashes"].values()))
        d["folder_size"] = sum(d["sizes"].values())
        d["date"] = get_time_str()
        d["legacy_version"] = legacy_version # if calculating hash will be changed or changed other things, this needed change too (increment)
        zfd.writestr("archive_info.json", json.dumps(d).encode("utf-8"))
    
    return (errors, )  # TODO: errors

def process_if_zip_exists(zip_path: str) -> int:
    """ 0=override,    1=skip """
    while(True):
        pout(f"File \"{zip_path}\" already exists. Override this?")
        pout("Type \"yes\" in capital letter to override or \"no\" to skip. \n> ", endl=False)
        user_in = input()
        if(user_in == "YES"):
            return 0
        if(user_in.lower().strip() == "no"):
            return 1
        pout(f"Wrong! Try again...")

def main_archive(args: list):

    parser = argparse.ArgumentParser(prog = "diwork archive",
        description="Makes archives from folders in the current directory. Make from each folder archive, where the folder and hash consist, and other data. ")
    parser.add_argument("folder_in", type=str, nargs=1,
                       help="Path to the directory where the folders from which you want to make an archive are located. ")
    parser.add_argument("folder_out", type=str, nargs=1,
                       help="Path to the directory where the archive will be placed. ")
    parser.add_argument("--yes", default=False, action='store_true',
                       help="No interractive. Answer \"yes\" always. ")
    parser.add_argument("--one", default=False, action='store_true',
                       help="Archive only one folder {folder_in}. Then {folder_in} path to the archived folder, not to directory of folders. ")

    parser = common_init_parser(parser)
    args = parser.parse_args(args)
    common_init_parse(args)

    IF_YES = args.yes
    IF_ONE = args.one

    root_dir = args.folder_in[0]
    if(is_folder(root_dir) == False):
        pout(f"\"{root_dir}\" is not directory. ")
        exit()
    root_dir = os.path.abspath(root_dir)

    out_dir = args.folder_out[0]
    if(is_folder(out_dir) == False):
        pout(f"\"{out_dir}\" is not directory. ")
        exit()
    out_dir = os.path.abspath(out_dir)

    err_out = []
    
    if(IF_ONE == False):
        folders = get_folders_one_deep(root_dir)
    else:
        folders = [root_dir]
    if(len(folders) == 0):
        pout(f"\"{folders}\" is empty. ")
        exit()
    folders = sorted([os.path.abspath(folder_i) for folder_i in folders])
    
    gi, gN = 0, len(folders)
    for folder_i in folders:
        gi-=-1
        if(IF_ONE == False):
            zip_path = os.path.join(out_dir, rel_path(folder_i, root_dir) + ".zip")
        else:
            zip_path = os.path.join(out_dir, os.path.basename(folder_i) + ".zip")
        pout(f"\t\t ({gi}/{gN}) Archiving folder \"{folder_i}\" to \"{zip_path}\"... ")
        if(is_exists(zip_path) == True):
            if(IF_YES == False):
                action = process_if_zip_exists(zip_path)
            else:
                pout(f"File \"{zip_path}\" already exists. It will be overridden! ")
                action = 0
            if(action == 1):
                pout("Skipping folder \"{folder_i}\". ")
                continue
            elif(action == 0):
                pout("File \"{zip_path}\" will be overridden. ")
            else:
                pout(f"Failed successfully. Unknown action={action}")
                exit()

        errors = make_archive_one_folder(folder_i, zip_path, f"({gi}/{gN})")[0]
        if(len(errors) > 0):
            err_out.append(f"\t *Error with folder \"{folder_i}\" BEGIN ")
            err_out += errors
            err_out.append(f"\t *Error with folder \"{folder_i}\"   END ")

    if(len(err_out) != 0):
        pout(f"\n===============\nSome troubles happened:")
        for err_i in err_out:
            pout(f"\t{err_i}")
        pout(f"===============")
    
    pout("\n=============== Done! ===============")



# -*- coding: utf-8 -*-

import os
import sys
import argparse 

from diwork_ways import *

class REP_OBJ():

    def __init__(self):
        self.IN_REP, self.OUT_REP = "{in}", "{out}" # absolute file path
        self.IN_REP_REL, self.OUT_REP_REL = "{in_rel}", "{out_rel}" # relative path
        self.IN_REP_FN, self.OUT_REP_FN = "{in_fn}", "{out_fn}" # file name with extension
        self.IN_REP_FNO, self.OUT_REP_FNO = "{in_fno}", "{out_fno}" # file name without extension
        self.IN_REP_FNE, self.OUT_REP_FNE = "{in_fne}", "{out_fne}" # file extension
        self.IN_REP_DIR, self.OUT_REP_DIR = "{in_dir}", "{out_dir}" # directory, where this file
        self.IN_REP_DIR_REL, self.OUT_REP_DIR_REL = "{in_dir_rel}", "{out_dir_rel}" # relative path of directory, where this file
        self.IN_REP_INDEX, self.OUT_REP_INDEX = "{in_i}", "{out_i}" # index, like 1,2,3...


def replace_REP_with_needed(command: str, REPS: list, REP_OBJ: object, file_abs_path: str, folder: str, index) -> str:
    file_rel_path = rel_path(file_abs_path, folder)
    file_fn = os.path.basename(file_abs_path)
    file_fno, file_fne = os.path.splitext(file_fn)
    if(file_fne != ""):
        file_fne = file_fne[1:]
    file_dir = os.path.dirname(file_abs_path)
    file_dir_rel = rel_path(file_dir, folder)
    for rep_i in REPS:
        if(rep_i in [REP_OBJ.IN_REP, REP_OBJ.OUT_REP]):
            command = command.replace(rep_i, file_abs_path)
        elif(rep_i in [REP_OBJ.IN_REP_REL, REP_OBJ.OUT_REP_REL]):
            command = command.replace(rep_i, file_rel_path)
        elif(rep_i in [REP_OBJ.IN_REP_FN, REP_OBJ.OUT_REP_FN]):
            command = command.replace(rep_i, file_fn)
        elif(rep_i in [REP_OBJ.IN_REP_FNO, REP_OBJ.OUT_REP_FNO]):
            command = command.replace(rep_i, file_fno)
        elif(rep_i in [REP_OBJ.IN_REP_FNE, REP_OBJ.OUT_REP_FNE]):
            command = command.replace(rep_i, file_fne)
        elif(rep_i in [REP_OBJ.IN_REP_DIR, REP_OBJ.OUT_REP_DIR]):
            command = command.replace(rep_i, file_dir)
        elif(rep_i in [REP_OBJ.IN_REP_DIR_REL, REP_OBJ.OUT_REP_DIR_REL]):
            command = command.replace(rep_i, file_dir_rel)
        elif(rep_i in [REP_OBJ.IN_REP_INDEX, REP_OBJ.OUT_REP_INDEX]):
            command = command.replace(rep_i, str(index))
        else:
            pout(f"replace_REP_with_needed ({rep_i}): Failed successfully")
            exit()
    return command

def help4command(repo) -> str:
    res = ""
    res += "Expected: \"python diwork.py exec {folder_in} {folder_out} {command}\", where: \n"
    res += "\t{command} like this \"convert \\\"{in}\\\" \\\"{out}\\\"\", where: \n"
    res += f"\t\t{repo.IN_REP}" + " each file from {folder_in} \n"
    res += f"\t\t{repo.OUT_REP}" + " each file from {folder_out} \n"
    res += "\t{folder_out} must be empty string (\"\") if do not used\n"
    res += "\n\tFor exampe, convert all jpg-pictures from \"path/to/jpg/image/folder\" to png-image:\n"
    res += "\t> python diwork.py exec \"path/to/jpg/image/folder\" \"path/to/png/image/folder\" \"convert \\\"{in}\\\" \\\"{out}.png\\\"\"\n"
    res += "\tTags for use:\n"
    res += f"\t\t{repo.IN_REP}" + ": full path of files from {folder_in}\n"
    res += f"\t\t{repo.OUT_REP}" + ": full path of files from {folder_out}\n"
    res += f"\t\t{repo.IN_REP_REL}" + ": relative path of files from {folder_in}\n"
    res += f"\t\t{repo.OUT_REP_REL}" + ": relative path of files from {folder_out}\n"
    res += f"\t\t{repo.IN_REP_FN}" + ": name of files from {folder_in}\n"
    res += f"\t\t{repo.OUT_REP_FN}" + ": name of files from {folder_out}\n"
    res += f"\t\t{repo.IN_REP_FNO}" + ": name without extension of files from {folder_in}\n"
    res += f"\t\t{repo.OUT_REP_FNO}" + ": name without extension of files from {folder_out}\n"
    res += f"\t\t{repo.IN_REP_FNE}" + ": files extension (without dot) from {folder_in}\n"
    res += f"\t\t{repo.OUT_REP_FNE}" + ": files extension (without dot) from {folder_out}\n"
    res += f"\t\t{repo.IN_REP_DIR}" + ": directory of files from {folder_in}\n"
    res += f"\t\t{repo.OUT_REP_DIR}" + ": directory of files from {folder_out}\n"
    res += f"\t\t{repo.IN_REP_DIR_REL}" + ": relative path of directory of files from {folder_in}\n"
    res += f"\t\t{repo.OUT_REP_DIR_REL}" + ": relative path of directory of files from {folder_out}\n"
    res += f"\t\t{repo.IN_REP_INDEX}" + ": index of files from {folder_in}\n"
    res += f"\t\t{repo.OUT_REP_INDEX}" + ": index of files from {folder_out}\n"
    res += "Example 1: \"" + repo.IN_REP + "\" and \"" + repo.IN_REP_DIR +"/" + repo.IN_REP_FNO + "." + repo.IN_REP_FNE + "\" the same text.\n"
    res += "Example 2: \"" + repo.OUT_REP_FN + "\" and \"" + repo.OUT_REP_FNO + "." + repo.OUT_REP_FNE +"\" the same text.\n"
    return res

def main_exec(args: list):
    repo = REP_OBJ()
    REPS_IN = [repo.IN_REP, repo.IN_REP_REL, repo.IN_REP_FN, repo.IN_REP_FNO, repo.IN_REP_FNE, repo.IN_REP_DIR, repo.IN_REP_DIR_REL, repo.IN_REP_INDEX]
    REPS_OUT = [repo.OUT_REP, repo.OUT_REP_REL, repo.OUT_REP_FN, repo.OUT_REP_FNO, repo.OUT_REP_FNE, repo.OUT_REP_DIR, repo.OUT_REP_DIR_REL, repo.OUT_REP_INDEX]

    parser = argparse.ArgumentParser(prog = "diwork exec",
        description="execute by certain rules. This module will help you execute a command with all files in certain directories.\n")
    parser.add_argument("folder_in", type=str, nargs=1,
                       help="Path to \"input\" directory")
    parser.add_argument("folder_out", type=str, nargs=1,
                       help="Path to \"output\" directory")
    parser.add_argument("command", type=str, nargs=1,
                       help=help4command(repo))
    parser.add_argument("--no_exec", default=False, action='store_true',
                       help="If True, the commands will not be executed, but only displayed. Default False")
    parser.add_argument("--exec_help", default=False, action='store_true',
                       help="If help looks terrible, then set this flag, like that: \"> python diwork.py exec "" "" "" --exec_help\"")
    parser = common_init_parser(parser)
    args = parser.parse_args(args)
    common_init_parse(args)


    if(args.exec_help == True):
        pout(help4command(repo))
        exit()
    folder_in = args.folder_in[0]
    folder_out = args.folder_out[0]
    command = args.command[0]
    NO_EXEC = args.no_exec

    err_out = []
    if(is_folder(folder_in) == False):
        pout(f"\"{folder_in}\" is not folder. ")
        exit()
    folder_in_abs = os.path.abspath(folder_in)
    if(folder_out != "" and is_folder(folder_out) == False):
        pout(f"\"{folder_out}\" is not folder. ")
        exit()
    if(folder_out == ""):
        #if OUT_REP in command:
        #    pout(f"{OUT_REP} in {command} finded, but folder_out is empty string. Exitting")
        #    exit()
        pass
    else:
        folder_out_abs = os.path.abspath(folder_out)

    if(folder_out != "" ):
        if(NO_EXEC == False):
            delete_all_if_dir_not_empty(folder_out_abs)

    if(NO_EXEC == True):
        print("===============GENERATED COMMANDS:===============")

    if(folder_out != ""):
        dirs_abs_in = list(set(getDirsList(folder_in_abs)))
        dirs_abs_in = sorted(dirs_abs_in)
        for dir_in_i in dirs_abs_in:
            dir_in_i_rel = rel_path(dir_in_i, folder_in_abs)
            dir_out_i_abs = os.path.join(folder_out_abs, dir_in_i_rel)
            cmd2exec = f"mkdir -p \"{dir_out_i_abs}\""
            if(NO_EXEC == False):
                # exe_out = exe(cmd2exec)
                # if(exe_out[1] != ""):
                #     pout(f"ERROR: {exe_out[1]}")
                #     exit()
                mkdir(dir_out_i_abs)
            else:
                pout(cmd2exec)
                

    files_abs_in = get_files_list(folder_in_abs)
    files_abs_in = sorted(files_abs_in)
    gi, N = 0, len(files_abs_in)
    for file_in_i in files_abs_in:
        gi+=1
        if(is_file(file_in_i) == False):
            err_out.append(f"\"{file_in_i}\" is not file or does not exists, it will be skipped. ")
            continue
        
        #command_i = command.replace(IN_REP, file_in_i)
        command_i = replace_REP_with_needed(command, REPS_IN, repo, file_in_i, folder_in_abs, gi)
        if(folder_out != ""):
            file_i_rel = rel_path(file_in_i, folder_in_abs)
            file_out_i = os.path.join(folder_out_abs, file_i_rel)
            #command_i = command_i.replace(OUT_REP, file_out_i)
            command_i = replace_REP_with_needed(command_i, REPS_OUT, repo, file_out_i, folder_out_abs, gi)
        if(NO_EXEC == False):
            pout(f"({gi}/{N}) Executing... ")
            #print(command_i)
            exe_out = exe(command_i)
            if(exe_out[1] != ""):
                err_out.append(f"ERROR: {exe_out[1]}")
                continue
        else:
            pout(command_i)

    os.sync()
    #exe("sync")

    if(len(err_out) != 0):
        pout(f"\n===============\nSome troubles happened:")
        for err_i in err_out:
            pout(f"\t{err_i}")
        pout(f"===============")

    pout("=============== Done! ===============")



# -*- coding: utf-8 -*-

import os
import sys
import argparse
import shutil

from diwork_ways import *

def main_hash(args: list) -> "list of hashes":

    parser = argparse.ArgumentParser(prog = "diwork hash",
        description="Calculate hash of directory(s)")
    parser.add_argument("folders_paths", type=str, nargs="+",
                       help="Paths to directories whose hash will be calculated")
    parser.add_argument("--exclude", type=str, nargs="+", default=None, action="append",
                       help="Do not take these files or directories into consideration when calculating the hash")
    # parser.add_argument("--symlink_mode", type=int, choices=[0,1,2], default=2, required=False,
    #                    help="What to do with links: 0 - ignor, 1 - consider link content, 2 - use the file where the link refers to. Default 2.")
    parser.add_argument("--hierarchy", default=False, action='store_true',
                       help="If True, then the file hierarchy will be considered when calculating the hash. Default False")
    parser = common_init_parser(parser)
    args = parser.parse_args(args)
    common_init_parse(args)

    folders = args.folders_paths
    files_exclude = args.exclude
    if(files_exclude != None):
        files_exclude = [shit[0] for shit in files_exclude]
        check_files_exists_or_exit(files_exclude)
    err_out = []
    folders_abs = [os.path.abspath(folder_i) for folder_i in folders]
    for folder_i in folders_abs:
        if(is_folder(folder_i) == False):
            pout(f"No such directory: \"{folder_i}\"")
            exit()
        if(folders_abs.count(folder_i) > 1):
            pout(f"Directory \"{folder_i}\" occurs several ({folders_abs.count(folder_i)}) times. Exiting...")
            exit()
    symlink_mode = args.symlink_mode
    if_hierarchy = args.hierarchy
    if(if_hierarchy == False):
        RES_OUT_PREFIX_HIERARCHY = "not"
    else:
        RES_OUT_PREFIX_HIERARCHY = "with"

    dir_hashes = {}
    ggi, ggN = 0, len(folders_abs)
    for folder_i in folders_abs:
        ggi-=-1
        pout(f"\nCalculating hash of directory \"{folder_i}\":")
        files = get_files_list(folder_i)
        files = sorted(files)
        files = exclude_files(files, files_exclude)
        hashes = []
        gi, files_len = 0, len(files)
        for file_i in files:
            gi+=1
            if(os.path.islink(file_i) == False and is_file(file_i) == False):
                err_out.append(f"\"{file_i}\" is not file or does not exists, it will be skipped. ")
                continue

            if(os.path.islink(file_i) == True):
                if(symlink_mode == 0):
                    pout(f"[{ggi}/{ggN}] ({gi}/{files_len}) \"{file_i}\" is symlink and symlink_mode={symlink_mode}. So it will be skipped. ")
                    continue
                elif(symlink_mode == 1):
                    linkto = os.readlink(file_i)
                    hash_i = get_hash_str(linkto)
                elif(symlink_mode == 2):
                    linkto = get_link_unwinding(file_i)
                    if(linkto == None):
                        pout(f"[{ggi}/{ggN}] ({gi}/{files_len}) \"{file_i}\" refers to nonexistent file. So it will be skipped. ")
                        err_out.append(f"\"{file_i}\" refers to nonexistent file, it will be skipped. ")
                        continue
                    file_i = linkto
                    hash_i = get_hash_file(file_i)
                else:
                    pout(f"Failed successfully. symlink_mode={symlink_mode}, cannot understand it. ")
                    exit()
            else:
                hash_i = get_hash_file(file_i)
            hashes.append(hash_i)
            pout(f"[{ggi}/{ggN}] ({gi}/{files_len}) Hash \"{hash_i}\" have file \"{file_i}\". ")
        hashes = sorted(hashes)

        # IF CHANGE, then change make_archive_one_folder in diwork_archive.py (its legacy_version)
        hash_files = get_hash_of_hashes(hashes)
        if(if_hierarchy == True):
            files_rel = [rel_path(fileh_i, folder_i) for fileh_i in files]
            dirs = getDirsList(folder_i)
            dirs_rel = [rel_path(dirh_i, folder_i) for dirh_i in dirs]
            hash_hierarchy = get_hash_of_hashes(  sorted(files_rel+dirs_rel) )
            hash_files = get_hash_str(hash_files + hash_hierarchy)

        pout(f"\n\nHash ({RES_OUT_PREFIX_HIERARCHY} considering the files hierarchy) of the directory \"{folder_i}\": \n==============================\n{hash_files}\n==============================\n")
        dir_hashes[folder_i] = hash_files


    if(len(err_out) != 0):
        pout(f"\n===============\nSome troubles happened:")
        for err_i in err_out:
            pout(f"\t{err_i}")
        pout(f"===============")

    col_1 = []
    col_2 = []
    for el in dir_hashes:
        col_1.append(dir_hashes[el])
        col_2.append(el)
    pout(f"\n\nHashes of directories ({RES_OUT_PREFIX_HIERARCHY} considering the files hierarchy): ")
    for i in range(len(col_1)):
        print(f"{col_1[i]} | {col_2[i]}")
    pout("\n=============== Done! ===============")

    return list(dir_hashes.values())



# -*- coding: utf-8 -*-

import os
import sys
import argparse 
import shutil

from diwork_ways import *

def define_clone_command() -> "str":
    if sys.platform == "linux" or sys.platform == "linux2" or sys.platform == "darwin":
        return "cp"
    elif sys.platform == "win32":  # In windows cp == copy
        return "copy"
    else:
        pout("(define_clone_command) Failed successfully. ")

def cp(src: str, dest: str, copy_mode: int = 0) -> None:
    if(copy_mode == 0):
        shutil.copy(src, dest)
    elif(copy_mode == 1):
        COPY_COMMAND = define_clone_command()
        exe_out = exe(f"{COPY_COMMAND} \"{src}\" \"{dest}\"")
        if(exe_out[1] != ""):
            pout(f"ERROR: {exe_out[1]}")
            exit()
    else:
        pout("(cp) Failed successfully. ")

def main_clone(args: list):
    parser = argparse.ArgumentParser(prog = "diwork clone",
        description="This module will clone all the contents of {folder_src} to {folder_dest}. As a result, {folder_src} and {folder_dest} will be completely identical.")
    parser.add_argument("folder_src", type=str, nargs=1,
                       help="Path to source directory")
    parser.add_argument("folder_dest", type=str, nargs=1,
                       help="Path to destination directory")
    # parser.add_argument("--symlink_mode", type=int, choices=[0,1,2], default=2, required=False,
    #                    help="What to do with links: 0 - ignor, 1 - consider link content, 2 - use the file where the link refers to. Default 2.")
    parser.add_argument("--copy_mode", type=int, choices=[0,1], default=0, required=False,
                       help="How to copy files: 0 - python shutil, 1 - system cp (copy). Default 0.")
    parser = common_init_parser(parser)
    args = parser.parse_args(args)
    common_init_parse(args)
    
    folder1 = args.folder_src[0]
    folder2 = args.folder_dest[0]
    err_out = []
    folder1_abs = os.path.abspath(folder1)
    folder2_abs = os.path.abspath(folder2)
    if(is_folder(folder1_abs) == False):
        pout(f"\"{folder1_abs}\" is not folder. ")
        exit()
    if(is_folder(folder2_abs) == False):
        pout(f"\"{folder2_abs}\" is not folder. ")
        exit()
    
    if(folder1_abs in folder2_abs):
        pout(f"Directory \"{folder1_abs}\" contains directory \"{folder2_abs}\". Exiting...")
        input("Enter to continue...")
    if(folder2_abs in folder1_abs):
        pout(f"Directory \"{folder2_abs}\" contains directory \"{folder1_abs}\". Exiting...")
        input("Enter to continue...")
    symlink_mode = args.symlink_mode
    copy_mode = args.copy_mode
    
    delete_all_if_dir_not_empty(folder2_abs)
    pout("Clonning...")

    dirs_abs_1 = getDirsList(folder1_abs)
    dirs_abs_1 = sorted(dirs_abs_1)
    for dir_i_1 in dirs_abs_1:
        dir_i_rel = rel_path(dir_i_1, folder1_abs)
        dir_i_2 = os.path.join(folder2_abs, dir_i_rel)
        mkdir(dir_i_2)
        #exe_out = exe(f"mkdir -p \"{dir_i_2}\"")
        #if(exe_out[1] != ""):
        #    pout(f"ERROR: {exe_out[1]}")
        #    exit()

    files_abs_1 = get_files_list(folder1_abs)
    files_abs_1 = sorted(files_abs_1)
    gi, N = 0, len(files_abs_1)
    for file_i_1 in files_abs_1:
        gi+=1
        if(os.path.islink(file_i_1) == False and is_file(file_i_1) == False):
            err_out.append(f"\"{file_i_1}\" is not file or does not exists, it will be skipped. ")
            continue

        file_i_rel = rel_path(file_i_1, folder1_abs)
        file_i_2 = os.path.join(folder2_abs, file_i_rel)
        if(os.path.islink(file_i_1) == True):
            if(symlink_mode == 0):
                pout(f"({gi}/{N}) \"{file_i_1}\" is symlink and symlink_mode={symlink_mode}. So it will be skipped. ")
                continue
            elif(symlink_mode == 1):
                linkto = os.readlink(file_i_1)
                os.symlink(linkto, file_i_2)
            elif(symlink_mode == 2):
                linkto = get_link_unwinding(file_i_1)
                if(linkto == None):
                    pout(f"({gi}/{N}) \"{file_i_1}\" refers to nonexistent file. So it will be skipped. ")
                    err_out.append(f"\"{file_i_1}\" refers to nonexistent file, it will be skipped. ")
                    continue
                pout(f"({gi}/{N}) Copying \"{file_i_rel}\" (\"{file_i_1}\"=\"{linkto}\" -> \"{file_i_2}\")... ")
                cp(linkto, file_i_2, copy_mode)
            else:
                pout(f"Failed successfully. symlink_mode={symlink_mode}, cannot understand it. ")
                exit()
        else:
            pout(f"({gi}/{N}) Copying \"{file_i_rel}\"... ")
            cp(file_i_1, file_i_2, copy_mode)
    if(copy_mode == 0):
        if sys.platform != "win32":
            os.sync()
    elif(copy_mode == 1):
        if sys.platform != "win32":
            exe("sync")
    else:
        pass

    if(len(err_out) != 0):
        pout(f"\n===============\nSome troubles happened:")
        for err_i in err_out:
            pout(f"\t{err_i}")
        pout(f"===============")

    pout("=============== Done! ===============")
